"""
Simplified statistics
"""
from .stat import unv, usd, unv_lambda, poisson_dist, normalize, novar_mean, mean, noisy, no_dist

__all__ = ['unv',
           'usd',
           'unv_lambda', 'poisson_dist', 
	   'no_dist',
	   'normalize', 'novar_mean',
           'mean', 'noisy'
           ]



from multiprocessing import Process,Queue
from smpl.doc import append_doc 

def queued(q,f,*args,**kwargs):
    q.put(f(*args,**kwargs))

def res(a):
    """
        Parallel evaluation of the list generator from :func:`gen`

        Examples
        ========

        >>> def twice(x):
        ...     return x+x
        >>> for r in [calc(twice,i) for i in range(0,5)]:
        ...     print(res(r))
        0
        2
        4
        6
        8
        >>> res([calc(lambda x : x**3, i) for i in range(0,5)])
        [0, 1, 8, 27, 64]

        Return parallel executed values
    """
    if isinstance(a,list):
        return [next(k) for k in a]
    return next(a)
def gen(f,*args,**kwargs):
    """
        Generates parallel execution list generator
    """
    q = Queue()
    p = Process(target=queued,args=(q,f,*args),kwargs=kwargs)
    yield p.start()
    yield [q.get(),p.join()][0]

@append_doc(res)
def calc(f,*args,**kwargs):
    g=gen(f,*args,**kwargs)
    next(g)
    return g

def par(f,*args,**kwargs):
    """
        Parallel execution of f on each element of args and kwargs

        Examples
        ========

        >>> par(lambda x : x**2, range(0,5))
        [0, 1, 4, 9, 16]

    """
    return res([calc(f,*[args[k][i] for k in range(len(args))],**{k:v[i] for k,v in kwargs.items()}) for i in range(len(args[0]) if len(args)>0 else len(next(iter(kwargs.values()))))])
 
 
 
if __name__ == "__main__":
    import doctest
    doctest.testmod()
 


"""
Simplified python code documentation
"""
from .doc import append_doc, append_plot,append_str, insert_doc,insert_str, insert_eq,table
__all__ = ['append_doc', 'append_plot', 'append_str','insert_doc', 'insert_str','insert_eq', 'table']


import numpy as np
import scipy
import uncertainties.unumpy as unp

from smpl import doc


# allgemeine Fitfunktionen
@doc.insert_eq()
@doc.append_plot(4)
def const(x,m):
    '''m'''
    return (np.ones(np.shape(x))*m)

@doc.insert_eq()
@doc.append_plot(2)
def linear(x,m): # lineare Funktion mit f(x) = m * x
    '''mx'''
    return(m*x)

@doc.insert_eq()
@doc.append_plot(2,-1)
def line(x, a, b): # gerade mit = f(x) = m * x + b
    '''ax+b'''
    return (a*x + b)
#Gerade=line
#Line=line

@doc.insert_eq()
@doc.append_plot(3,0.02,3)
def cos_abs(x, a, f, phi):
    '''$a|\\cos(2πf(x-\\phi))|$'''
    return a * np.abs(unp.cos(2*np.pi*f*(x-phi)))

@doc.insert_eq()
@doc.append_plot(3,0.02,3)
def cos(x, a, f, phi):
    '''$a\\cos(2πf(x-\\phi))$'''
    return a * unp.cos(2*np.pi*f*(x-phi))
@doc.insert_eq()
@doc.append_plot(3,0.02,3)
def sin(x, a, f, phi):
    '''$a\\sin(2πf(x-\\phi))$'''
    return a * unp.sin(2*np.pi*f*(x-phi))
@doc.insert_eq()
@doc.append_plot(3,0.02,3)
def tan(x, a, f, phi):
    '''$\\tan(2πf(x-\\phi))a$'''
    return a * unp.tan(2*np.pi*f*(x-phi))

@doc.insert_eq()
@doc.append_plot(0,5,3,0)
def lorentz(x,x_0,A,d,y):
    '''$\\frac{A}{\\pi d (1+ (\\frac{x-x_0}{d})^2)} + y$'''
    return 1/(np.pi*d*(1+(x-x_0)**2/d**2))*A + y
#Lorentz=lorentz


@doc.insert_eq()
@doc.append_plot(0,5,3,0)
def gauss(x, x_0, A, d, y):
    '''$A\\cdot \\exp\\left(\\frac{-(x-x_0)^2}{2d^2}\\right)+y$'''
    return A * unp.exp(-(x - x_0)**2 / 2 / d**2) + y
Gauss=gauss


@doc.insert_eq()
@doc.append_plot(0.5,4)
def exponential(x, c, y_0):
    '''$\\exp(cx)y_0$'''
    return unp.exp(c * x) * y_0
exp=exponential

@doc.insert_eq()
@doc.append_plot(0.5,4,xmin=0.1)
def log(x, c, y_0):
    '''$\\log(cx)y_0$'''
    return unp.log(c * x) * y_0

@doc.insert_eq()
@doc.append_plot(1,5,0)
def square(x,x_0,A,y):
    '''$A(x-x_0)^2+y$'''
    return A*(x-x_0)**2+y
def quadratic_(x,A,y):
    '''A*(x)**2+y'''
    return A*(x)**2+y
quadratic=square
Quadratisch=quadratic_

@doc.insert_eq()
@doc.append_plot(1,5,0,-2)
def cube(x,a,b,c,d):
    '''$ax^3+bx^2+cx+d$'''
    return a*x**3+b*x**2+c*x+d


@doc.insert_eq()
@doc.append_plot(1,3.3,0)
def order(x,A,k,y):
    '''A*(x)**k+y'''
    return A*(x)**k+y


@doc.insert_eq()
@doc.append_plot(1,3.3,0,xmin=0)
def sqrt(x,a,b,c):
    '''$a\\sqrt{x+b}+c$'''
    return a*unp.sqrt(x+b)+c


def Gauß(x, x0, A, d):
    '''$A\\cdot \\exp\\left(\\frac{-(x-x0)^2}{2d^2}\\right)$'''
    return A * unp.exp(-(x - x0)**2 / 2 / d**2)
def Two_Gauss(x, x0, A0, d0, x1, A1, d1,y):
    '''$A\\cdot \\exp\\left(\\frac{-(x-x0)^2}{2d^2}\\right)+y$'''
    #return A0 * np.exp(-(x - x0)**2 / 2 / d0**2) + y0 + A1 * np.exp(-(x - x1)**2 / 2 / d1**2) + y1
    return gauss(x,x0,A0,d0,y)+gauss(x,x1,A1,d1,0)
def Six_Gauss(x, x0, A0, d0, x1, A1, d1,x2, A2, d2,x3, A3, d3,x4, A4, d4, x5, A5, d5,y):
    '''$A\\cdot \\exp\\left(\\frac{-(x-x0)^2}{2d^2}\\right)+y$'''
    #return A0 * np.exp(-(x - x0)**2 / 2 / d0**2) + y0 + A1 * np.exp(-(x - x1)**2 / 2 / d1**2) + y1
    return gauss(x,x0,A0,d0,y)+gauss(x,x1,A1,d1,0)+gauss(x,x2,A2,d2,0)+gauss(x,x3,A3,d3,0)+gauss(x,x4,A4,d4,0)+gauss(x,x5,A5,d5,0)
def Two_Exp(x,A0,A1,l0,l1):
    '''A0*exp(-l0*x)+A1*exp(-l1*x)'''
    return exponential(x,-l0,A0)+exponential(x,-l1,A1)

def Two_Lorentz(x, x0, A0, d0, x1, A1, d1,y):
    '''$\\frac{A}{\\pi d (1+ (\\frac{x-x0}{d})^2)} + y$'''
    return lorentz(x,x0,A0,d0,y)+lorentz(x,x1,A1,d1,0)

def Split_Gauss(x,x0,A0,d0,d1,y):
    '''\n$A\\cdot \\exp\\left(\\frac{-(x-x0)^2}{2d0^2}\\right)+y$, für x>x0\n$A\\cdot \\exp\\left(\\frac{-(x-x0)^2}{2d1^2}\\right)+y$, sonst'''
    return np.where(x>x0,gauss(x,x0,A0,d0,y) ,gauss(x,x0,A0,d1,y))


from sympy.parsing.sympy_parser import standard_transformations,implicit_multiplication_application
from sympy.parsing.sympy_parser import parse_expr
import sympy
from sympy.printing.pycode import pycode
import uncertainties.unumpy as unp

import numpy as np

def get_lambda(expr,xvar):
    if isinstance(expr,str):
        return str_get_lambda(expr,xvar)
    else:
        return fnc_get_lambda(expr,xvar)

def fnc_get_lambda(expr,xvar):
    __l__ = eval("lambda " + ','.join(get_varnames(expr,xvar)) + ": expr(" + ','.join(expr.__code__.co_varnames) + ")",{"expr":expr})
    __l__.__doc__ = expr.__doc__
    return __l__

def str_get_lambda(expr,xvar):
    parsed_expr = str_get_expr(expr)
    pc = pycode(parsed_expr)
    for s in unp.__all__:
        pc = pc.replace("math." + s, "unp."+s)
    # old direct way. Doesn't use unp.
    #__l__ = sympy.lambdify(get_varnames(expr),parsed_expr)
    __l__ = eval("lambda " + ','.join(get_varnames(expr,xvar)) + ": "+ pc)
    #exec("global __l__; __l__ = lambda " + ','.join(get_varnames(expr)) + ": "+ pc)
    return __l__

def get_varnames(expr,xvar):
    if isinstance(expr,str):
        return str_get_varnames(expr,xvar)
    else:
        return fnc_get_varnames(expr,xvar)
#def get_argcount(expr):
#    return len(get_varnames(expr))

def str_get_varnames(expr,xvar):
    """
    Returns a list of variables used in the ``str`` math-expression via sympy.
    """

    parsed_expr = str_get_expr(expr)
    
    new_locals = [sym.name
                  for sym in parsed_expr.atoms(sympy.Symbol)]
    new_locals = sorted(new_locals)
    if xvar is None:
        return np.roll(new_locals,1)
    else:
        vars = new_locals
        assert(xvar in vars)
        vars.remove(xvar)
        return [xvar,*vars]


def fnc_get_varnames(func,xvar):
    if xvar is None:
        return func.__code__.co_varnames
    else:
        vars = [s for s in func.__code__.co_varnames]
        assert(xvar in vars)
        vars.remove(xvar)
        return [xvar,*vars]

def str_get_expr(expr):
    expr = expr.replace("unp.","").replace("np.","")
    parsed_expr = sympy.parsing.sympy_parser.parse_expr(
        expr,
        local_dict=None,
        transformations=(standard_transformations + (implicit_multiplication_application,)),
        evaluate=False
    )
    return parsed_expr


import pathlib
import glob
import pathlib
#from io import StringIO
from smpl import debug
import numpy as np
import os
import sys
from smpl.io import mkdirs

def gf(i):
    """
    Scientific format with ``i`` digits.
    """
    return "{0:." + str(i) + "g}"

def si(s,u="",fmt="{}"):
    """
    Get number with uncertainty and unit in ``si`` format for latex.

    Parameters
    ==========
    s : ufloat
        number to be returned in a latex compatible format
    u : str
        unit of that number
    fmt : str
        format string for the numbers

    Returns
    =======
    sistr : str
        latex SI string of the number with it's uncertainty and unit.
    """
    return "\\SI{%s}{%s}"%((fmt.format(s)).replace("/","").replace("(","").replace(")",""),u)

def si_line(a,skip = 0,fmt="{}"):
    """
    Get array ``a`` in the format of a line of a latex table.
    """
    return si_tab(np.transpose([[t] for t in a]),skip,fmt)
def si_ttab(tab,skip=0, fmt="{}"):
    """
    Transposed :func:`si_tab`.
    """
    return si_tab(np.transpose(tab),skip,fmt)
def si_tab(tab,skip=0, fmt="{}"):
    """
    Get arrays of (uncertainty) numbers in  a latex table compatible form.

    Parameters
    ==========
    tab : array_like
        Array containing the values of the table
    skip : number
        Skip this many table lines
    fmt : str
        format string for the numbers

    Returns
    =======
    tabstr : str
        table latex string
    """
    #mkdirs(fn)
    #file = open(fn,"w")
    s = ""
    for i in range(len(tab)):
        for j in range(len(tab[i])):
            if(j!=0):
                s += "&"
            if(j>=skip):
                s+=si(tab[i][j],fmt=fmt)
            else:
                s+="%s"%(tab[i][j])
        s += "\\\\\n"
    return s




"""
A list of various functions
"""
from .functions import const, lorentz,linear,line,cos_abs,gauss,square,cube,exp,log,sqrt,tan,cos,sin
__all__ = [
    'const', 
    'lorentz',
    'cos',
    'sin',
    'tan',
    'linear',
    'line',
    'cos_abs',
    'gauss',
    'square',
    'cube',
    'sqrt',
    'exp',
    'log'
    ]


import matplotlib.patches as mpatches
import numpy as np
import statistics as stat
import scipy as sci
import scipy.integrate as integrate
import scipy.fftpack
import matplotlib.pyplot as plt
import matplotlib
import matplotlib.axes as axes
from matplotlib import colors as mcolors
import math
from scipy import optimize
import uncertainties as unc
import uncertainties.unumpy as unp
import uncertainties.umath as umath
import glob
import os
import matplotlib.pyplot as plt
from scipy.odr import *
import sympy
# from tqdm import tqdm
import matplotlib.pylab as pylab
import pathlib
import types
import inspect
# local imports
from smpl import functions
from smpl import io
from smpl import util
from smpl import wrap
from smpl.doc import append_doc, append_str
from smpl import doc
from smpl import fit as ffit
import inspect


def set_plot_style():
    # fig_size = (8, 6)
    # fig_legendsize = 14
    # fig_labelsize = 12 # ‘xx-small’, ‘x-small’, ‘small’, ‘medium’, ‘large’, ‘x-large’, ‘xx-large’.
    params = {'legend.fontsize': 'x-large',
              'figure.figsize': (8, 6),
              'axes.labelsize': 'x-large',
              'axes.titlesize': 'x-large',
              'xtick.labelsize': 'x-large',
              'ytick.labelsize': 'x-large'}
    pylab.rcParams.update(params)
    matplotlib.rcParams.update(params)
    # matplotlib.rcParams.update({'font.size': fig_labelsize})
    # colors = dict(mcolors.BASE_COLORS, **mcolors.CSS4_COLORS)


unv = unp.nominal_values
usd = unp.std_devs

default = {
    #    'params'        :[None      ,"Initial fit parameters",
    'xlabel': ["", "X axis label", ],
    'ylabel': ["", "Y axis label", ],
    'label': [None, "Legend name of plotted ``data``", ],
    'fmt': ['.', "Format for plotting fit function", ],
    'units': [None, "Units of the fit parameters as strings. Displayed in the Legend", ],
    'save': [None, " File to save the plot", ],
    'lpos': [0, "Legend position", ],
    #          'frange'        :[None      ,"Limit the fit to given range. First integer is the lowest and second the highest index.",],
    'prange': [None, "Limit the plot of the fit to given range", ],
    'sigmas': [0, "Color the array of given ``sigma`` times uncertainty. Only works if the fit function is coded with ``unp``", ],
    'data_sigmas': [1, "Color the array of given ``sigma`` times uncertainty. Only works if the data has uncertainties", ],
    'init': [False, "Initialize a new plot"],
    'ss': [True, "save, add legends and grid to the plot", ],
    'also_data': [True, " also plot the data"],
    'also_fit': [True, "also plot the fit", ],
    'logy': [False, "logarithmic x axis", ],
    'logx': [False, "logarithmic y axis", ],
    'data_color': [None, "Color of the data plot", ],
    'fit_color': [None, "Color of the fit plot", ],
    'residue': [False, "Display difference between fit and data in a second plot", ],
    'residue_err': [True, "Differences between fit and data will have errorbars", ],
    'show': [False, "Call plt.show()", ],
    'size': [None, "Size of the plot as a tuple (x,y). Only has an effect if ``init`` is True", ],
    'number_format': [io.gf(4), "Format to display numbers.", ],
    # ,          'selector'      :[ None     ,"Function that takes ``x`` and ``y`` as parameters and returns an array mask in order to limit the data points for fitting. Alternatively a mask for selecting elements from datax and datay.",],
    # ,          'fixed_params'  :[ True     ,"Enable fixing parameters by choosing the same-named variables from ``kwargs``.",],
    # ,          'sortbyx'       :[ True     , "Enable sorting the x and y data so that x is sorted.",],
    'interpolate': [True, "Enable interpolation of whole data if fit range is limited by ``frange`` or ``selector``.", ],
    'interpolate_min': [None, "Lower interpolation bound", ],
    'interpolate_max': [None, "Higher interpolation bound", ],
    'interpolate_hatch': [r"||", "Interpolation shape/hatch for filled area in case of ``sigmas``>0. See https://matplotlib.org/stable/gallery/shapes_and_collections/hatch_style_reference.html", ],
    'bbox_to_anchor': [None, "Position in a tuple (x,y),Shift position of the legend out of the main pane. ", ],
    'ncol': [None, "Columns in the legend if used with ``bbox_to_anchor``.", ],
    'steps': [1000, "resolution of the plotted function", ],
    'fitinline': [False,  "No newlines for each fit parameter", ],
    'grid': [True,  "Enable grid for the plot", ],
    'hist': [False, "Enable histogram plot", ],
    'stairs': [False, "Enable stair plot", ],

}


# @doc.insert_str("\tDefault kwargs\n\n\t")
@doc.append_doc(ffit.fit_kwargs)
@doc.append_str(doc.table(default))
@doc.append_str(doc.table({"plot_kwargs        ": ["default", "description"]}, bottom=False))
def plot_kwargs(kwargs):
    """
    Set default plot_kwargs if not set.
    """
    kwargs = ffit.fit_kwargs(kwargs)
    for k, v in default.items():
        if not k in kwargs:
            kwargs[k] = v[0]
    return kwargs


# @append_doc(default_kwargs)
def auto(datax, datay, funcs=None, **kwargs):
    """
    Automatically loop over functions and fit the best one.

    Parameters
    ==========
    funcs : function array
        functions to consider as fit. Default all ``smpl.functions``.
    **kwargs : optional
        see :func:`plot_kwargs`.

    Returns
    =======
    The best fit function and it's parameters. Also a lambda function where the parameters are already applied.



    """
    best_f, best_ff, lambda_f = ffit.auto(datax, datay, funcs, **kwargs)
    if not best_f is None:
        fit(datax, datay, best_f, **kwargs)
    return best_f, best_ff, lambda_f


# @append_doc(default_kwargs)
def fit(datax, datay, function, **kwargs):
    """Fit and plot function to datax and datay.

    Parameters
    ==========
    datax : array_like
        X data either as ``unp.uarray`` or ``np.array`` or ``list``
    datay : array_like
        Y data either as ``unp.uarray`` or ``np.array`` or ``list``
    function : func
        Fit function with parameters: ``x``, ``params``
    **kwargs : optional
        see :func:`plot_kwargs`.
    Fit parameters can be fixed via ``kwargs`` eg. ``a=5``.

    Returns
    =======
    array_like
        Optimized fit parameters of ``function`` to ``datax`` and ``datay``

    Examples
    ========

    .. plot::
        :include-source:

        >>> from smpl import functions as f
        >>> from smpl import plot
        >>> param = plot.fit([0,1,2],[0,1,2],f.line)
        >>> plot.unv(param).round()[0]
        1.0

    """
    kwargs = plot_kwargs(kwargs)
    fit = None
    fig = None
    fig = init_plot(**kwargs)
    if kwargs['also_data']:
        plt_data(datax, datay, **kwargs)
    if kwargs['also_fit']:
        fit, kwargs['fit_color'] = plt_fit(datax, datay, function, **kwargs)
    if kwargs['ss']:
        kwargs['oldshow'] = kwargs['show']
        kwargs['show'] = kwargs['show'] and not kwargs['residue']
        save_plot(**kwargs)
        kwargs['show'] = kwargs['oldshow']
    if kwargs['residue'] and fig is not None:
        plt_residue(datax, datay, function, fit, fig, **kwargs)
    return fit

# @append_doc(default_kwargs)


def data(datax, datay, function=None, **kwargs):
    """Plot datay against datax via :func:`fit`

    Parameters
    ==========
    datax : array_like
        X data either as ``unp.uarray`` or ``np.array`` or ``list``
    datay : array_like
        Y data either as ``unp.uarray`` or ``np.array`` or ``list``
    function : func,optional
        Fit function with parameters: ``x``, ``params``
    **kwargs : optional
        see :func:`plot_kwargs`.
    Returns
    =======
    array_like
        Optimized fit parameters of ``function`` to ``datax`` and ``datay``
    """
    if not 'also_fit' in kwargs:
        kwargs['also_fit'] = False
    kwargs = plot_kwargs(kwargs)
    if kwargs['label'] == None and kwargs['lpos'] == 0:
        kwargs['lpos'] = -1
    return fit(datax, datay, function, **kwargs)


def _plot(x, y, **kwargs):
    args = [x, y]
    kargs = {}
    if util.has('fmt', kwargs):
        args += [kwargs["fmt"]]
    if util.has('label', kwargs) and kwargs['label'] != "":
        kargs['label'] = kwargs['label']
    if util.has('color', kwargs) and kwargs['color'] != "":
        kargs['color'] = kwargs['color']
    plt.plot(*args, **kargs)
# @append_doc(default_kwargs)


def function(func, *args, **kwargs):
    """
    Plot function ``func`` between ``xmin`` and ``xmax``

    Parameters
    ==========
    func : function
        Function to be plotted between ``xmin`` and ``xmax``, only taking `array_like` ``x`` as parameter
    *args : optional
        arguments for ``func``
    **kwargs : optional
        see :func:`plot_kwargs`.
    """
    if not util.has("xmin", kwargs) or not util.has("xmin", kwargs):
        raise Exception("xmin or xmax missing.")

    if not util.has('lpos', kwargs) and not util.has('label', kwargs):
        kwargs['lpos'] = -1
    if not util.has('fmt', kwargs):
        kwargs['fmt'] = "-"

    kwargs = plot_kwargs(kwargs)
    xfit = np.linspace(kwargs['xmin'], kwargs['xmax'], kwargs['steps'])
    fig = init_plot(**kwargs)

    if not util.has("label", kwargs) or kwargs['label'] is None:
        kwargs['label'] = get_fnc_legend(func, args, **kwargs)
        kwargs['lpos'] = 0
    _plot(xfit, func(xfit, *args), **kwargs)
    if kwargs['ss']:
        save_plot(**kwargs)


# xaxis="",yaxis="",fit_color=None,save = None,residue_err=True,show=False):
def plt_residue(datax, datay, gfunction, fit, fig, **kwargs):
    function = wrap.get_lambda(gfunction, kwargs['xvar'])
    frame2 = fig.add_axes((.1, .1, .8, .2))
    kwargs['yaxis'] = "$\\Delta$" + kwargs['yaxis']
    kwargs['data_color'] = kwargs['fit_color']

    if kwargs['residue_err']:
        plt_data(datax, datay-function(datax, *fit), **kwargs)
    else:
        plt_data(unv(datax), unv(datay-function(datax, *fit)), **kwargs)
    kwargs['lpos'] = -1
    save_plot(**kwargs)


def data_split(datax, datay, **kwargs):
    return ffit.data_split(datax, datay, **kwargs)


def _fit(datax, datay, function, **kwargs):
    """
    Returns a fit like :func:`fit` but does no plotting.
    """
    return ffit.fit(datax, datay, function, **kwargs)

# @append_doc(plot_kwargs)


# xaxis="",yaxis="",label=None,fmt=None,data_color=None):
def plt_data(datax, datay, **kwargs):
    """
        Plot datay vs datax
    """
    x, y, xerr, yerr = data_split(datax, datay, **kwargs)
    if xerr is not None:
        xerr = xerr * kwargs['data_sigmas']
    if yerr is not None:
        yerr = yerr * kwargs['data_sigmas']

    if xerr is None and yerr is None:
        if kwargs['fmt'] is None:
            plt.plot(x, y, label=kwargs['label'], color=kwargs['data_color'])
        elif kwargs['fmt'] == "step":
            plt.step(x, y, where='mid',
                     label=kwargs['label'], color=kwargs['data_color'])
        elif kwargs['fmt'] == "hist":
            plt.step(x, y, where='mid',
                     label=kwargs['label'], color=kwargs['data_color'])
            plt.fill_between(x, y, step="mid")
        else:
            plt.plot(x, y, kwargs['fmt'], label=kwargs['label'],
                     color=kwargs['data_color'])
    else:
        if kwargs['fmt'] is None:
            plt.errorbar(x, y, yerr=yerr, xerr=xerr, fmt=" ", capsize=5,
                         label=kwargs['label'], color=kwargs['data_color'])
        elif kwargs['fmt'] == "step":
            ll, = plt.step(x, y, where='mid',
                           color=kwargs['data_color'])
            if xerr is not None:
                for ix in range(len(x)):
                    dx = (xerr[ix])
                    tx = [x[ix]-dx, x[ix]+dx]
                    plt.fill_between(tx, y[ix]-yerr[ix], y[ix]+yerr[ix],
                                     label=kwargs['label']if ix == 1 else None, alpha=0.2, step='pre', color=ll.get_color())
            else:
                plt.fill_between(x, y-yerr, y+yerr,
                                 label=kwargs['label'], alpha=0.2, step='mid', color=ll.get_color())
        elif kwargs['fmt'] == "hist":
            plt.errorbar(x, y, yerr=yerr, xerr=xerr, fmt=" ", capsize=5,
                         color="black")
            plt.fill_between(x, y, step="mid", label=kwargs['label'])
        else:
            plt.errorbar(x, y, yerr=yerr, xerr=xerr, fmt=kwargs['fmt'], capsize=5,
                         label=kwargs['label'], color=kwargs['data_color'])


def get_fnc_legend(function, fit, **kwargs):
    if isinstance(function, str):
        l = "$" + sympy.latex(wrap.str_get_expr(function)) + "$"
    else:
        l = function.__name__
    # l = ""
    if l == "<lambda>":
        # l = "$\\lambda$(" +  ','.join(function.__code__.co_varnames) + ") = " #sympy.latex(eval(function.__code__.co_code))
        try:
            cc, li = inspect.findsource(function)
            f = ''.join(cc[li:]).split('lambda')[-1].split(':')[1].split(',')[
                0].replace("\n", "")  # .replace("#", "").replace("'''","")
            l = "$" + sympy.latex(wrap.str_get_expr(f)) + "$"
        except OSError:
            l = "$\\lambda$(" + ','.join(function.__code__.co_varnames) + ")"

    if function.__doc__ is not None and not isinstance(function, str):
        l = function.__doc__.split('\n')[0]
    vnames = wrap.get_varnames(function, kwargs['xvar'])
    for i in range(1, len(vnames)):
        l = l + ("\n" if not kwargs["fitinline"] or i == 1 else " ")
        l = l + "$" + sympy.latex(sympy.symbols(str(vnames[i]))) + "$="
        if kwargs['units'] is not None and usd(fit[i-1]) > 0:
            l = l + "("
        if 'number_format' in kwargs:
            l = l + kwargs['number_format'].format(fit[i-1])
        else:
            l = l + "%s" % (fit[i-1])

        if kwargs['units'] is not None and usd(fit[i-1]) > 0:
            l = l + ")"
        if kwargs['units'] is not None:
            l = l + " " + kwargs['units'][i-1]
    return l

# @append_doc(default_kwargs)


# p0=None,units=None,frange=None,prange=None,sigmas=1,residue=False, fig = None,fit_color=None):
def plt_fit(datax, datay, gfunction, **kwargs):
    """
       Plot Fit 
    """
    function = wrap.get_lambda(gfunction, kwargs['xvar'])
    x, y, xerr, yerr = data_split(datax, datay, **kwargs)
    fit = _fit(datax, datay, gfunction, **kwargs)
    if kwargs['prange'] is None:
        xfit = np.linspace(np.min(unv(x)), np.max(unv(x)), 1000)
    else:
        xfit = np.linspace(kwargs['prange'][0], kwargs['prange'][1], 1000)
    l = get_fnc_legend(gfunction, fit, **kwargs)
    ll = None
    if kwargs['sigmas'] > 0:
        ll, = plt.plot(xfit, function(xfit, *unv(fit)),
                       "-", color=kwargs['fit_color'])
        yfit = function(xfit, *fit)
        plt.fill_between(xfit, unv(yfit)-kwargs['sigmas']*usd(yfit), unv(
            yfit)+kwargs['sigmas']*usd(yfit), alpha=0.4, label=l, color=ll.get_color())
    else:
        ll, = plt.plot(xfit, function(xfit, *unv(fit)), "-",
                       label=l, color=kwargs['fit_color'])
    if (kwargs['frange'] is not None or kwargs['selector'] is not None) and util.true('interpolate', kwargs) or util.has("interpolate_max", kwargs) or util.has("interpolate_min", kwargs):
        xxfit = np.linspace(util.get("interpolate_min", kwargs, np.min(
            unv(datax))), util.get("interpolate_max", kwargs, np.max(unv(datax))))
        plt.plot(xxfit, unv(function(xxfit, *fit)), "--", color=ll.get_color())
        if kwargs['sigmas'] > 0:
            xxxfit = np.linspace(np.min(xxfit), np.min(xfit))
            yfit = function(xxxfit, *fit)
            plt.fill_between(xxxfit, unv(yfit)-kwargs['sigmas']*usd(yfit), unv(yfit)+kwargs['sigmas']*usd(
                yfit), alpha=0.4, color=ll.get_color(), hatch=util.get("interpolate_hatch", kwargs, r"||"))
            xxxfit = np.linspace(np.max(xfit), np.max(xxfit))
            yfit = function(xxxfit, *fit)
            plt.fill_between(xxxfit, unv(yfit)-kwargs['sigmas']*usd(yfit), unv(yfit)+kwargs['sigmas']*usd(
                yfit), alpha=0.4, color=ll.get_color(), hatch=util.get("interpolate_hatch", kwargs, r"||"))
    return fit, ll.get_color()


def init_plot(**kwargs):  # size=None,residue=False): #init
    # fig = plt.figure(figsize=fig_size)
    fig = None
    if kwargs['init'] or util.true("residue", kwargs):
        if kwargs['size'] is None:
            fig = plt.figure()
        else:
            fig = plt.figure(figsize=kwargs['size'])
        if kwargs['residue']:
            frame1 = fig.add_axes((.1, .3, .8, .6))
    if util.has("xlabel", kwargs) and kwargs['xlabel'] != "":
        plt.xlabel(kwargs['xlabel'])
    if util.has("ylabel", kwargs) and kwargs['ylabel'] != "":
        plt.ylabel(kwargs['ylabel'])
    if util.has("xaxis", kwargs) and kwargs['xaxis'] != "":
        plt.xlabel(kwargs['xaxis'])
    if util.has("yaxis", kwargs) and kwargs['yaxis'] != "":
        plt.ylabel(kwargs['yaxis'])
    return fig


def save_plot(**kwargs):  # save=None,lpos=0,logy=False,logx=False,show=True): #save
    """
        save plot
    """
    if 'logy' in kwargs and kwargs['logy']:
        plt.gca().set_yscale('log')
    if 'logx' in kwargs and kwargs['logx']:
        plt.gca().set_xscale('log')
    plt.tight_layout()
    if 'lpos' in kwargs and kwargs['lpos'] >= 0:
        if(util.has('bbox_to_anchor', kwargs)):
            if(util.has('ncol', kwargs)):
                plt.legend(loc=kwargs['lpos'], bbox_to_anchor=kwargs['bbox_to_anchor'],
                           ncol=kwargs['ncol'], borderaxespad=0)
            else:
                plt.legend(loc=kwargs['lpos'],
                           bbox_to_anchor=kwargs['bbox_to_anchor'])
        else:
            plt.legend(loc=kwargs['lpos'])
    # plt.gca().set_xlim([kwargs['xmin'],kwargs['xmax']])
    # plt.gca().set_ylim([kwargs['ymin'],kwargs['ymax']])
    if 'save' in kwargs and not kwargs['save'] == None:
        io.mkdirs(kwargs['save'])
        plt.savefig(kwargs['save'] + ".pdf")
    plt.grid(b=kwargs["grid"])
    if 'show' in kwargs and kwargs['show']:
        show(**kwargs)


def show(**kwargs):
    kwargs = plot_kwargs(kwargs)

    plt.grid(b=kwargs["grid"])
    plt.show()


if __name__ == "__main__":
    import doctest
    doctest.testmod()
